#include "BioAlignment.h"
#include <iomanip>
#include <sstream>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdexcept>
#include <algorithm>

using namespace std;
namespace bio {

   BioAlignment::ScoreMatrix_t BioAlignment::matrix;
   string BioAlignment::lastLoadedMatrix;
   pthread_mutex_t BioAlignment::matrixLoadMutex = PTHREAD_MUTEX_INITIALIZER;

   /** No descriptions */
   BioAlignment::BioAlignment() {
   }
   /** No descriptions */
   BioAlignment::BioAlignment(const BioSequence &sCon, const BioSequence &sSeq) {
      assign(sCon, sSeq);
   }
   /** No descriptions */
   BioAlignment::BioAlignment(const BioSubSequence &sCon, const BioSubSequence &sSeq) {
      assign(sCon, sSeq);
   }
   /** No descriptions */
   BioAlignment::BioAlignment(const BioAlignment& other) :
      m_sCon(other.m_sCon), m_sSeq(other.m_sSeq), Score(other.Score) {
   }
   /** No descriptions */
   BioAlignment::~BioAlignment() {
   }
   /** No descriptions */
   void BioAlignment::clear() {
      m_sCon.clear();
      m_sSeq.clear();
   }
   //--------------------------------------
   BioAlignment& BioAlignment::assign(const BioSequence& sCon, const BioSequence& sSeq) {
      m_sCon.assign(sCon);
      m_sSeq.assign(sSeq);
      return *this;
   }
   //--------------------------------------
   BioAlignment& BioAlignment::assign(const BioSubSequence& sCon, const BioSubSequence& sSeq) {
      m_sCon.assign(sCon);
      m_sSeq.assign(sSeq);
      return *this;
   }
   //--------------------------------------
   BioAlignment& BioAlignment::assign(const BioAlignment& aln) {
      setScore(aln.getScore());
      return assign(aln.getCon(), aln.getSeq());
   }
   //--------------------------------------
   BioSubSequence& BioAlignment::getCon() {
      return m_sCon;
   }
   //--------------------------------------
   BioSubSequence& BioAlignment::getSeq() {
      return m_sSeq;
   }
   //--------------------------------------
   //double BioAlignment::calcSim(bool ignoreCpG) const
   double BioAlignment::calcSim() const {
      return BioAlignment::calcSim(m_sCon, m_sSeq, &masks);
   }
   //--------------------------------------
   void BioAlignment::printHead(ostream& os) const {
      printAlnHead(os);
      os << endl;
   }
   //--------------------------------------
   void BioAlignment::printAlnHead(ostream& os) const {
      AlnStat stat = getStat();
      m_sCon.guessType();
      m_sSeq.guessType();
      os << setw(12) << setiosflags(ios::left) << m_sSeq.getName() << " ";
      os << setw(7) << resetiosflags(ios::left) << min(m_sSeq.getFrom(), m_sSeq.getTo()) << " ";
      os << setw(7) << setiosflags(ios::left) << max(m_sSeq.getFrom(), m_sSeq.getTo()) << " ";
      os << setw(12) << resetiosflags(ios::left) << m_sCon.getName() << " ";
      os << setw(7) << min(m_sCon.getFrom(), m_sCon.getTo()) << " ";
      os << setw(7) << setiosflags(ios::left) << max(m_sCon.getFrom(), m_sCon.getTo());
      os << ((m_sSeq.getTo() - m_sSeq.getFrom()) * (m_sCon.getTo() - m_sCon.getFrom()) >= 0 ? " d  " : " c  ");
      os << " " << setw(5) << setiosflags(ios::left) << setprecision(4) << calcSim();
      //      <<(stat.iMCnt+stat.iMMCnt > 0)?stat.iMCnt/(stat.iMCnt+stat.iMMCnt):0;
      os << " " << setw(5) << setprecision(2) << ((m_sCon.getseqtype() == NA) && (m_sSeq.getseqtype() == NA)) ? (double) stat.iMMCnt
            / (stat.iTRCnt + 1) : ((stat.iMCnt + stat.iMMCnt > 0) ? (stat.iPCnt / (stat.iMCnt + stat.iMMCnt)) : 0);
      os << " " << setw(5) << resetiosflags(ios::left) << (int) getScore();
   }
   //--------------------------------------
   void BioAlignment::print(ostream& os, unsigned int iLineLen) const {
      printAlnHead(os);
      os << " <\n\n";
      printAln(os, iLineLen);
   }
   //======================================
   int BioAlignment::read(istream& is) {
      return m_sCon.read(is) && m_sSeq.read(is);
   }
   //======================================
   int BioAlignment::readPaln(istream& is) throw () {
      /*
       top: AluJo  1 - 283     bot: AluMacYa3  1 - 283

       1        10        20        30        40        50        60        70
       |2 4 6 8 | 2 4 6 8 | 2 4 6 8 | 2 4 6 8 | 2 4 6 8 | 2 4 6 8 | 2 4 6 8 |
       GGCCGGGCGCGGTGGCTCACGCCTGTAATCCCAGCACTTTGGGAGGCCGAGGCGGGAGGATTGCTTGAGC
       ******************* *******************************|**** ****||*  ***
       GGCCGGGCGCGGTGGCTCAAGCCTGTAATCCCAGCACTTTGGGAGGCCGAGACGGGCGGATCAC--GAGG

       71       80        90        100       110        120        130
       |2 4 6 8 | 2 4 6 8 | 2 4 6 8 | 2 4 6 8 | 2 4 6 8  | 2 4 6 8  | 2 4 6 8
       CCAGGAGTTCGAGACCAGCCTGGGCAACATAGCGAGACCCCGTCTCTAC-AAAAAATAC-AAAAATTAGC
       |****** ********* ***** |****||*|**|************* ********* *****|****
       TCAGGAGATCGAGACCATCCTGGCTAACACGGTGAAACCCCGTCTCTACTAAAAAATACAAAAAACTAGC

       AluJo      -  AluMacYa3: m= 120  mm= 16  t= 8   g= 3  s= 0.87 r= 2.00 n=  95.8
       */
      clear();
      string line;
      char name1[64];
      char name2[64];
      char sDummy[64];
      unsigned int p11, p12, p21, p22, m, mm, t, g;
      string s1, s2, s3;
      while (!getline(is, line, '\n').fail()) {
         if (line.find("top: ") == 0) {
            BioSequence::shrink(line, ' ');
            if (sscanf(line.c_str(), "top: %s %d - %d bot: %s %d - %d", name1, &p11, &p12, name2, &p21, &p22) == 6) {
               m_sSeq.setName(name1);
               m_sCon.setName(name2);
               m_sSeq.setFrom(p11);
               m_sSeq.setTo(p12);
               m_sCon.setFrom(p21);
               m_sCon.setTo(p22);
               while (true) {
                  for (int i = 0; i < 3; i++) {
                     if (getline(is, line, '\n').fail()) {
                        cerr << "ERROR: Unexpected end of file\n";
                     }
                     if (line.find(name1) == 0) {
                        BioSequence::shrink(line, ' ');
                        if (sscanf(line.c_str(), "%s - %s m= %d mm= %d t= %d g= %d", sDummy, sDummy, &m, &mm, &t, &g) == 6) {
                           Score = m * 9 - mm * 12 + t * 6 - g * 30 - (m_sCon.size() - m - mm - g) * 6;
                           return m_sCon.size();
                        } else {
                           cerr << "ERROR: Failed to read paln tail line:\n";
                           cerr << line << "\n";
                           return 0;
                        }
                     }
                  }
                  if (getline(is, s1, '\n').fail()) {
                     cerr << "ERROR: Unexpected end of file\n";
                     return 0;
                  }
                  if (getline(is, s2, '\n').fail()) {
                     cerr << "ERROR: Unexpected end of file\n";
                     return 0;
                  }
                  if (getline(is, s3, '\n').fail()) {
                     cerr << "ERROR: Unexpected end of file\n";
                     return 0;
                  }
                  m_sSeq.append(s1);
                  m_sCon.append(s3);
               }
            } else {
               cerr << "ERROR: Failed to read header of paln aoutput:\n";
               cerr << line << "\n";
               return 0;
            }
         }
      }
      return 0;
   }
   //======================================
   int BioAlignment::readAxt(istream& is) throw () {
      clear();
      string line;
      do {
         getline(is, line, '\n');
      } while (line[0] == '#');
      //#Example of axt format
      //0 chr19 3001012 3001075 chr11 70568380 70568443 - 3500
      //TCAGCTCATAAATCACCTCCTGCCACAAGCCTGGCCTGGTCCCAGGAGAGTGTCCAGGCTCAGA
      //TCTGTTCATAAACCACCTGCCATGACAAGCCTGGCCTGTTCCCAAGACAATGTCCAGGCTCAGA
      //
      unsigned int id;
      char name1[64];
      char name2[64];
      char dir;
      unsigned int p11, p12, p21, p22, score;
      if (sscanf(line.c_str(), "%d %s %d %d %s %d %d %c %d", &id, name1, &p11, &p12, name2, &p21, &p22, &dir, &score) == 9) {
         getline(is, getCon(), '\n');
         getline(is, getSeq(), '\n');
         getline(is, line, '\n');
         getCon().setName(name1);
         getCon().setFrom(p11);
         getCon().setTo(p12);
         getSeq().setName(name2);
         if (dir == '+') {
            getSeq().setFrom(p21);
            getSeq().setTo(p22);
         } else {
            getSeq().setFrom(p22);
            getSeq().setTo(p21);
         }
         setScore(score);
         return getCon().size();
      } else {
         return -1;
      }
   }
   //======================================
   ostream& operator <<(ostream& os, const BioAlignment& aln) {
      aln.print(os);
      return os;
   }
   /** Read property of double Score. */
   const double& BioAlignment::getScore() const {
      return Score;
   }
   /** Write property of double Score. */
   void BioAlignment::setScore(const double& _newVal) {
      Score = _newVal;
   }
   /** No descriptions */
   const AlnStat BioAlignment::getStat() const {
      return getStat(NULL);
   }
   /**
    * Calculates similarity in alignment
    */
   const AlnStat BioAlignment::getStat(vector<string> *ignoreList) const {
      return getStat(getCon(), getSeq(), ignoreList);
   }
   AlnStat BioAlignment::getStat(const BioSequence& seq1, const BioSequence& seq2, const vector<string> *ignoreList) {
      string::const_iterator cptr, sptr;
      AlnStat stat;
      stat.reset();
      unsigned int iCGapLen = 0;
      unsigned int iSGapLen = 0;
      SEQTYPE iType;
      if ((iType = seq1.guessType()) == SEQTYPE_UNKNOWN) {
         iType = seq2.guessType();
      }
      for (cptr = seq1.begin(), sptr = seq2.begin(); cptr != seq1.end() && sptr != seq2.end(); cptr++, sptr++) {
         if (*cptr == ScoreMatrix_t::GAPCHAR && *sptr == ScoreMatrix_t::GAPCHAR) {
            continue;
         }
         if (*cptr == ScoreMatrix_t::LPADDING || *sptr == ScoreMatrix_t::LPADDING || *cptr == ScoreMatrix_t::RPADDING || *sptr
               == ScoreMatrix_t::RPADDING) {
            continue;
         }
         if (ignoreList != NULL && ignoreList->size() > 0) {
            bool skip = false;
            for (vector<string>::const_iterator vptr = ignoreList->begin(); vptr != ignoreList->end(); vptr++) {
               string::const_iterator sptr1, sptr2;
               int gaplen = 0;
               int gapsum = 0;
               for (sptr1 = cptr, sptr2 = (*vptr).begin(); sptr1 != seq1.end() && sptr2 != (*vptr).end(); sptr1 = seq1.next_base(sptr1,
                                                                                                                                 &gaplen), sptr2++) {
                  if (toupper(*sptr1) != toupper(*sptr2)) {
                     break;
                  }
                  gapsum += gaplen;
               }
               if (sptr2 == (*vptr).end()) {
                  skip = true;
                  for (int i = (*vptr).size() - 1; i > 0; i--) {
                     cptr++;
                     sptr++;
                     cptr += gapsum;
                     sptr += gapsum;
                  }
                  break;
               }
            }
            if (skip) {
               continue;
            }
         }

         if (*cptr != *sptr) {
            if (*cptr == ScoreMatrix_t::GAPCHAR) {
               iCGapLen++;
               continue;
            }
            if (*sptr == ScoreMatrix_t::GAPCHAR) {
               iSGapLen++;
               continue;
            }
         } else {
            if (*cptr == ScoreMatrix_t::GAPCHAR)
               continue;
         }
         if (iCGapLen > 0) {
            stat.iMMCnt++;
            stat.iGCnt++;
            iCGapLen = 0;
         }
         if (iSGapLen > 0) {
            stat.iMMCnt++;
            stat.iGCnt++;
            iSGapLen = 0;
         }
         if (matrix.getScore(*cptr, *sptr) > 0) {
            stat.iPCnt++;
         }
         if ((iType == AA ? isAAMatch(*cptr, *sptr) : isNAMatch(*cptr, *sptr))) {
            stat.iMCnt++;
         } else {
            stat.iMMCnt++;
            if (iType == NA) {
               if (isTransition(*cptr, *sptr)) {
                  stat.iTRCnt++;
               } else {
                  stat.iTVCnt++;
               }
            }
         }
      }
      return stat;
   }
   /** No descriptions */
   bool BioAlignment::isTransition(char c1, char c2) {
      c1 = toupper(c1);
      c2 = toupper(c2);
      switch (c1) {
      case 'A':
         return (c2 == 'G');
      case 'T':
         return (c2 == 'C');
      case 'C':
         return (c2 == 'T');
      case 'G':
         return (c2 == 'A');
      }
      return false;
   }
   /** No descriptions */
   bool BioAlignment::isNAMatch(char c1, char c2) {
      if (isalpha(c1)) {
         c1 = toupper(c1);
         c2 = toupper(c2);
         if (c1 == c2 && c1 != 'N' && c1 != 'X') {
            return true;
         }
      }
      return false;
   }
   /** No descriptions */
   bool BioAlignment::isAAMatch(char c1, char c2) {
      if (isalpha(c1) || c1 == '*') {
         c1 = toupper(c1);
         c2 = toupper(c2);
         if (c1 == c2 && c1 != 'X') {
            return true;
         }
      }
      return false;
   }
   /** No descriptions */
   const BioSubSequence& BioAlignment::getCon() const {
      return m_sCon;
   }
   /** No descriptions */
   const BioSubSequence& BioAlignment::getSeq() const {
      return m_sSeq;
   }
   /** No descriptions */
   char BioAlignment::getLink(char c1, char c2) const {
      if (m_sCon.getseqtype() == NA && m_sSeq.getseqtype() == NA) {
         if (isNAMatch(c1, c2)) {
            return '|';
         }
         if (isTransition(c1, c2)) {
            return ':';
         }
      } else {
         if (isAAMatch(c1, c2)) {
            return '|';
         }
      }
      if (matrix.getScore(c1, c2) > 0) {
         return '+';
      }
      if (isalpha(c1) && isalpha(c2)) {
         return ' ';
      }
      return '-';
   }
   /** No descriptions */
   void BioAlignment::addMask(string s) {
      if (s.length() > 0) {
         masks.push_back(s);
      }
   }
   /** No descriptions */
   void BioAlignment::clearMasks() {
      masks.clear();
   }
   /** No descriptions */
   void BioAlignment::printAln(ostream& os, int iLineLen, const char* sBlahS, const char* sBlahQ) const {
      string sCon;
      string sSeq;
      string::iterator cPtr, sPtr;
      unsigned long iPos;
      for (iPos = 0; iPos < m_sCon.size(); iPos += iLineLen) {
         sCon = m_sCon.substr(iPos, iLineLen);
         sSeq = m_sSeq.substr(iPos, iLineLen);
         os << "S: " << sSeq << sBlahS << "\n   ";
         for (cPtr = sCon.begin(), sPtr = sSeq.begin(); cPtr != sCon.end() && sPtr != sSeq.end(); cPtr++, sPtr++) {
            os << getLink(*cPtr, *sPtr);
         }
         os << "\nQ: " << sCon << sBlahQ << "\n\n";
      }
   }
   /**  */
   void BioAlignment::loadMatrix(const char* s) {
      assert(pthread_mutex_lock(&matrixLoadMutex) == 0);
      if (lastLoadedMatrix.compare(s) != 0) {
         if (matrix.parse(s)) {
            lastLoadedMatrix = s;
         }
      }
      pthread_mutex_unlock(&matrixLoadMutex);
   }

   const BioAlignment::ScoreMatrix_t& BioAlignment::getMatrix() {
      return matrix;
   }
   /**  */
   //===========================================================
   char BioAlignment::getAvg(string& s, int mode) {
      // mode:
      // 0 - default, all symbols are accounted;
      // 1 - gaps are ignored;
      // 2 - only gaps and non-gaps (N) are distinguished
      char arr[255];
      int max = 0;
      char res = '-';
      string::iterator sptr;
      for (int i = 0; i < 255; i++) {
         arr[i] = 0;
      }
      for (sptr = s.begin(); sptr != s.end(); sptr++) {
         char c = toupper(*sptr);
         switch (mode) {
         case 1:
            if (isalpha(c) && c != 'N' && c != 'X') {
               arr[c]++;
               if (max < arr[c]) {
                  max = arr[c];
                  res = c;
               }
            }
            break;
         case 2:
            if (c != '-') {
               c = 'N';
            }
         default:
            arr[c]++;
            if (max < arr[c]) {
               max = arr[c];
               res = c;
            }
         }
      }
      return res;
   }

   /**
    * Specialized function to remove indels from pairwise alignment.
    * Optionaly it saves removed insertions into os1, os2 ostreams.
    */
   void BioAlignment::ungap(int minGap, ostream *os1, ostream *os2) throw () {
      if (m_sSeq.size() != m_sCon.size()) {
         cerr << "ERROR: Bad alignment - different length of sequences.";
         return;
      }
      string sSeq = m_sSeq;
      string sCon = m_sCon;
      string seqSlider;
      string conSlider;
      if ((int(minGap / 2)) * 2 == minGap) {
         minGap++;
      }
      int avgWin = 1 + 2 * int(minGap / 2);
      int size = sSeq.size();
      int i;
      for (i = size - 1, seqSlider = m_sSeq.substr(max(0, i - avgWin), 1 + 2 * avgWin), conSlider = m_sCon.substr(max(0, i - avgWin), 1 + 2
            * avgWin); i >= 0; i--, seqSlider = m_sSeq.substr(max(0, i - avgWin), 1 + 2 * avgWin), conSlider = m_sCon.substr(max(0, i
            - avgWin), 1 + 2 * avgWin)) {
         if (getAvg(seqSlider, 2) == '-') {
            sSeq[i] = '-';
         } else {
            //            sSeq[i] = 'N';
         }
         if (getAvg(conSlider, 2) == '-') {
            sCon[i] = '-';
         } else {
            //            sCon[i] = 'N';
         }
      }
      //    cerr << sSeq << "\n\n";
      //    cerr << sCon << "\n\n";

      //    int seqRmCnt = 0;
      //    int conRmCnt = 0;
      //    int rmCnt = 0;
      int p11, p12, p21, p22;
      p11 = p12 = p21 = p22 = 0;
      for (i = size - 1, seqSlider = sSeq.substr(max(0, i - minGap), 1 + 2 * minGap), conSlider = sCon.substr(max(0, i - minGap), 1 + 2
            * minGap); i >= 0; i--, seqSlider = sSeq.substr(max(0, i - minGap), 1 + 2 * minGap), conSlider
            = sCon.substr(max(0, i - minGap), 1 + 2 * minGap)) {
         if (getAvg(conSlider, 2) == '-' || getAvg(seqSlider, 2) == '-') {
            if (isalpha(m_sSeq[i])) {
               if (isalpha(m_sCon[i])) {
                  //                    cerr << "ERROR: Code error: deleting non-gap position.\n";
                  //                    throw(baddata_exception("Code error: deleting non-gap position."));
               } else {
                  //                seqRmCnt++;
                  if (p12 == 0) {
                     p12 = i;
                     p11 = i;
                  } else {
                     p11 = i;
                  }
               }
            }
            if (isalpha(m_sCon[i])) {
               if (isalpha(m_sSeq[i])) {
                  //                    cerr << "ERROR: Code error: deleting non-gap position.\n";
                  //                    throw(baddata_exception("Code error: deleting non-gap position."));
               } else {
                  //                conRmCnt++;
                  if (p22 == 0) {
                     p22 = i;
                     p21 = i;
                  } else {
                     p21 = i;
                  }
               }
            }
            //            rmCnt++;
         } else {
            if (p11 + p12 > 0 && p12 - p11 + 1 >= minGap) {
               if (os1) {
                  (*os1) << ">" << m_sSeq.getName() << "_" << m_sSeq.getFrom() << "-" << m_sSeq.getTo() << "[" << p11 + 1 << "-" << p12 + 1
                        << "]\n";
                  (*os1) << m_sSeq.substr(p11, p12 - p11 + 1) << "\n";
               }
               m_sSeq.erase(p11, p12 - p11 + 1);
               m_sCon.erase(p11, p12 - p11 + 1);
            }
            if (p21 + p22 > 0 && p22 - p21 + 1 >= minGap) {
               if (os2) {
                  (*os2) << ">" << m_sCon.getName() << "_" << m_sCon.getFrom() << "-" << m_sCon.getTo() << "[" << p21 + 1 << "-" << p22 + 1
                        << "]\n";
                  (*os2) << m_sCon.substr(p21, p22 - p21 + 1) << "\n";
               }
               m_sSeq.erase(p21, p22 - p21 + 1);
               m_sCon.erase(p21, p22 - p21 + 1);
            }
            p21 = p22 = 0;
            p11 = p12 = 0;
         }
      }
      //    m_sSeq.setTo(m_sSeq.getTo()-seqRmCnt);
      //    m_sCon.setTo(m_sCon.getTo()-conRmCnt);
   }
   bool BioAlignment::isSelfHit() {
      if (m_sCon.getName() == m_sSeq.getName() && m_sCon.getFrom() == m_sSeq.getFrom() && m_sCon.getTo() == m_sSeq.getTo()) {
         return true;
      } else {
         return false;
      }
   }

   ostream& operator <<(ostream& os, const AlnStat& stat) {
      stat.print(os);
      return os;
   }

   //--------------------------------------
   double BioAlignment::calcSim(const BioSequence& seq1, const BioSequence& seq2, const vector<string>* masks) {
      long iCGapLen = 0;
      long iSGapLen = 0;
      int iMMCnt = 0;
      int iMCnt = 0;
      SEQTYPE iType;
      if ((iType = seq1.guessType()) == SEQTYPE_UNKNOWN) {
         iType = seq2.guessType();
      }
      for (unsigned int i = 0; i < seq1.size() && i < seq2.size(); i++) {
         if (masks != NULL && masks->size() > 0) {
            bool match = 0;
            for (vector<string>::const_iterator vptr = masks->begin(); vptr != masks->end(); vptr++) {
               unsigned int ci, mi;
               for (ci = i, mi = 0; ci < seq1.size() && mi < (*vptr).size(); ci++, mi++) {
                  if (toupper(seq1[ci]) != toupper((*vptr)[mi])) {
                     break;
                  }
               }
               if (mi == (*vptr).size()) {
                  match = 1;
                  i += ((*vptr).size() - 1);
                  break;
               }
            }
            if (match) {
               continue;
            }
         }
         if (seq1[i] == ScoreMatrix_t::GAPCHAR && seq2[i] == ScoreMatrix_t::GAPCHAR) {
            continue;
         }
         if (seq1[i] == ScoreMatrix_t::GAPCHAR) {
            iCGapLen++;
            continue;
         }
         if (seq2[i] == ScoreMatrix_t::GAPCHAR) {
            iSGapLen++;
            continue;
         }
         if (iCGapLen > 0) {
            iMMCnt++;
            iCGapLen = 0;
         }
         if (iSGapLen > 0) {
            iMMCnt++;
            iSGapLen = 0;
         }
         if ((iType == AA ? isAAMatch(seq1[i], seq2[i]) : isNAMatch(seq1[i], seq2[i]))) {
            iMCnt++;
         } else {
            iMMCnt++;
         }
      }
      return (iMCnt + iMMCnt > 0) ? double(iMCnt) / (iMCnt + iMMCnt) : 0;
   }
   double BioAlignment::calcScore() const {
      return calcScore(0, getCon().size());
   }
   double BioAlignment::calcScore(size_t start, size_t end) const {
      return calcScore(getCon(), getSeq(), start, end, NULL, false, false);
   }
   double BioAlignment::calcScore(const string& s1, const string& s2, size_t start, size_t end, const ScoreMatrix_t* matptr,
                                  vector<double>* columnScores, bool cumulative, bool ignoreGaps) {
      const ScoreMatrix_t& matrix = matptr == NULL ? BioAlignment::matrix : *matptr;
      if (s1.size() != s2.size()) {
         throw invalid_argument("Sequences are of different sizes");
      }
      size_t size = s1.size();
      if (start < 0) {
         start = 0;
      }
      if (end >= size) {
         end = size - 1;
      }
      if (!cumulative) {
         if (columnScores != NULL) {
            columnScores->assign(size, 0);
         }
      }
      double sumScore = 0;
      double rowScore = 0;
      double colScore = 0;
      bool isInGap = false;
      for (int i = start; i <= end; i++) {
         rowScore = colScore = 0;
         if (s1[i] == ScoreMatrix_t::GAPCHAR || s2[i] == ScoreMatrix_t::GAPCHAR) {
            if (!ignoreGaps) {
               colScore = matrix.getScore(s1[i], s2[i]);
               if (s1[i] != s2[i]) {
                  if (isInGap) {
                     rowScore = matrix.getGapExtPenalty();
                  } else {
                     rowScore = matrix.getGapIniPenalty();
                     isInGap = true;
                  }
               } else {
                  isInGap = false;
               }
            }
         } else {
            isInGap = false;
            rowScore = colScore = matrix.getScore(s1[i], s2[i]);
         }
         sumScore += rowScore;
         if (columnScores != NULL) {
            (*columnScores)[i] += colScore;
         }
      }
      return sumScore;
   }

   BioAlignment& BioAlignment::sharp(int edgeWidht) {
      if (edgeWidht <= 0) {
         return *this;
      }
      int aln1 = -1;
      int aln2 = -1;
      int edge = 0;
      int start = 0;
      for (int i = 0; i < getCon().size(); i++) {
         if (getCon()[i] == getSeq()[i]) {
            if (isalpha(getCon()[i]) || getCon()[i] == STOPCODONSYMB) {
               edge++;
            }
         } else {
            edge = 0;
            start = i + 1;
         }
         if (edge >= edgeWidht) {
            aln1 = start;
            break;
         }
      }
      trimLeft(aln1);

      edge = 0;
      start = getCon().size() - 1;
      for (int i = start; i >= 0; i--) {
         if (getCon()[i] == getSeq()[i]) {
            if (isalpha(getCon()[i]) || getCon()[i] == STOPCODONSYMB) {
               edge++;
            }
         } else {
            edge = 0;
            start = i - 1;
         }
         if (edge >= edgeWidht) {
            aln2 = start;
            break;
         }
      }
      trimRight(getCon().size() - aln2 - 1);

      return *this;
   }

   BioAlignment& BioAlignment::trimLeft(size_t trimLength) {
      if (trimLength >= getCon().size()) {
         setScore(0);
         getCon().clear();
         getSeq().clear();
         return *this;
      }
      int refDelta = getCon().getFrom() <= getCon().getTo() ? 1 : -1;
      int seqDelta = getSeq().getFrom() <= getSeq().getTo() ? 1 : -1;

      for (size_t i = 0; i < trimLength && i < m_sCon.size(); i++) {
         if (isalpha(getCon()[i]) || getCon()[i] == STOPCODONSYMB) {
            getCon().setFrom(getCon().getFrom() + refDelta);
         }
         if (isalpha(getSeq()[i]) || getSeq()[i] == STOPCODONSYMB) {
            getSeq().setFrom(getSeq().getFrom() + seqDelta);
         }
      }
      setScore(calcScore() - calcScore(0, trimLength - 1));
      getCon().erase(0, trimLength);
      getSeq().erase(0, trimLength);
      return *this;
   }

   BioAlignment& BioAlignment::trimRight(size_t trimLength) {
      if (trimLength >= getCon().size()) {
         setScore(0);
         getCon().clear();
         getSeq().clear();
         return *this;
      }
      int refDelta = getCon().getFrom() <= getCon().getTo() ? 1 : -1;
      int seqDelta = getSeq().getFrom() <= getSeq().getTo() ? 1 : -1;

      for (size_t x = 0; x < trimLength && x < m_sCon.size(); x++) {
         int i = getCon().size() - 1 - x;
         if (isalpha(getCon()[i]) || getCon()[i] == STOPCODONSYMB) {
            getCon().setTo(getCon().getTo() - refDelta);
         }
         if (isalpha(getSeq()[i]) || getSeq()[i] == STOPCODONSYMB) {
            getSeq().setTo(getSeq().getTo() - seqDelta);
         }
      }
      setScore(calcScore() - calcScore(getCon().size() - trimLength, getCon().size() - 1));
      getCon().erase(getCon().size() - trimLength, trimLength);
      getSeq().erase(getSeq().size() - trimLength, trimLength);
      return *this;
   }
}
