/*
 * concurrent_set.hpp
 *
 *  Created on: Feb 22, 2010
 *      Author: kohany
 */

#ifndef CONCURRENT_SET_HPP_
#define CONCURRENT_SET_HPP_

#include <set>
#include <stdexcept>
#include <pthread.h>

using namespace std;

template <class T>
class concurrent_set {
   typedef set<T> set_t;
   set_t set_;
   pthread_rwlock_t rwlock;
   void inline init_rwlock() {
      if (pthread_rwlock_init(&rwlock, NULL)) {
         throw(runtime_error("Failed to initialize rwlock"));
      }
   }
public:
   typedef typename set<T>::iterator iterator;
   typedef typename set<T>::const_iterator const_iterator;
   typedef typename set<T>::value_type value_type;
   concurrent_set() {
      init_rwlock();
   }
   ~concurrent_set() {
      pthread_rwlock_destroy(&rwlock);
   }
   pair<const_iterator,bool> insert ( const value_type& x );
   size_t erase(const T& v);
};

template <class T>
pair<typename concurrent_set<T>::const_iterator,bool> concurrent_set<T>::insert ( const value_type& v ) {
   if (pthread_rwlock_rdlock(&rwlock)) {
      throw(runtime_error("Failed to lock rwlock for reading"));
   }
   //typename set_t::const_iterator sptr(set_.find(v));
   pair<typename concurrent_set<T>::const_iterator,bool> res(set_.find(v), false);
   pthread_rwlock_unlock(&rwlock);
   if (res.first == set_.end()) {
      if (pthread_rwlock_wrlock(&rwlock)) {
         throw(runtime_error("Failed to lock rwlock for writing"));
      }
      try {
         res = set_.insert(v);
      } catch (...) {
         pthread_rwlock_unlock(&rwlock);
         throw ;
      }
      pthread_rwlock_unlock(&rwlock);
   }
   return res;
}

template <class T>
size_t concurrent_set<T>::erase(const T& v) {
   size_t rv;
   if (pthread_rwlock_wrlock(&rwlock)) {
      throw(runtime_error("Failed to lock rwlock for writing"));
   }
   try {
      rv = set_.erase(v);
   }
   catch(...) {
      pthread_rwlock_unlock(&rwlock);
      throw;
   }
   pthread_rwlock_unlock(&rwlock);
   return rv;
}

#endif /* CONCURRENT_SET_HPP_ */
