/***************************************************************************
                          cseqstat.cpp  -  description
                             -------------------
    begin                : Fri Dec 6 2002
    copyright            : (C) 2002 by Oleksiy Kohany
    email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cseqstat.h"
#include "constants.h"
#include <ctype.h>
#include <string.h>
#include <iostream>
#include <sstream>
#include <iomanip>

namespace bio {

#define STOPCODON '*'
const char VOIDCHAR[] = "NXnx";
using namespace std;

bool CSeqStat::sbWithN = 0;
/**==============================================================*/
CSeqStat::CSeqStat(){
   reset();
}
/**==============================================================*/
CSeqStat::CSeqStat(string& sName){
   Name = sName;
   reset();
}
/**==============================================================*/
CSeqStat::~CSeqStat(){
}
/** Read property of string Name. */
const string& CSeqStat::getName() const {
   return Name;
}
/** Write property of string Name. */
void CSeqStat::setName( const string& _newVal){
   Name = _newVal;
}
/** No descriptions */
void CSeqStat::addData(const char* pcLine) {
   const char* pcPtr;
   for(pcPtr = pcLine; pcPtr != 0 && *pcPtr != '\0'; pcPtr++) {
      if(*pcPtr < BASERANGE) {
        Stat[toupper(*pcPtr)]++;
        if(isalpha(*pcPtr)) {
          Size++;
        }
      }
      else {
         throw(out_of_range("Encountered out of range symbol"));
      }
   }
}
/** No descriptions */
void CSeqStat::addData(const string& sLine) {
   addData(sLine.c_str());
}
/** No descriptions */
void CSeqStat::reset(){
   for(int i = 0; i < BASERANGE; i++) {
      Stat[i] = 0;
   }
   Size = 0;
}

/** No descriptions */
SEQTYPE CSeqStat::guessType() const {
   if(Stat['A'] + Stat['T'] + Stat['C'] + Stat['G'] + Stat['U'] > 0) {
      if((double)(Stat['A'] + Stat['T'] + Stat['C'] + Stat['G'] + Stat['U']) / (Size - Stat['N'] - Stat['X']) > 0.75) {
         return NA;
      }
   }
   else {
      if(((double)(Stat['N'])) / (Size) > 0.75) {
         return NA;
      }
   }
   return AA;
}
/** No descriptions */
void CSeqStat::print(ostream* os) const {
    SEQTYPE type = guessType();
   *os<<getName()<<"\t"<<typeName(type)<<"\t"<<getLength()<<"\t"<<Size;
   if(type == NA) {
      (*os) << "\t" << setw(6) << setprecision(4);
      if(Stat['C'] + Stat['G']) {
        *os << ((double) Stat['C'] + Stat['G']) / (Stat['A'] + Stat['T'] + Stat['C'] + Stat['G']);
      }
      else {
        *os << 0;
      }
   }
   (*os) << "\n";
}
/** No descriptions */
void CSeqStat::print(FILE* os) const {
//   fprintf(os, "%-15s %-3s %15lu", getName().c_str(), typeName(guessType()), getLength());
   fprintf(os, "%s\t%s\t%lu\t%lu", getName().c_str(), typeName(guessType()), getLength(), Size);
   if(guessType() == NA) {  //If it is nucleic acid, print C+G content
      fprintf(os, "\t%6.4f",
      (Stat['C'] + Stat['G'] == 0) ? 0 : ((double)Stat['C'] + Stat['G']) / (Stat['A'] + Stat['T'] + Stat['C'] + Stat['G']));
   }
   fprintf(os, "\n");
}
/** No descriptions */
const char* CSeqStat::typeName(SEQTYPE type) {
   static char ssName[MAXNAMESIZE];
   switch (type) {
      case NA:
         strcpy(ssName, "n");
         break;
      case AA:
         strcpy(ssName, "p");
         break;
      case SEQTYPE_UNKNOWN:
         default:
         strcpy(ssName, "x");
   }
   return ssName;
}
/** Reads and analyses next sequences from the stream. */
int CSeqStat::read(istream& is){
    BioSequence seq;
    reset();
    if(seq.read(is) >= 0) {
        setName(seq.getName());
        addData(seq);
        return Size;
    }
    return -1;
}

ostream& operator << (ostream& os, const CSeqStat& stat) {
	stat.print(&os);
  return os;
}

/** No descriptions */
CSeqStat& CSeqStat::operator += (const CSeqStat& sst){
   for(int i = 0; i < BASERANGE; i++) {
      Stat[i] += sst.getStat()[i];
      if(isalpha(toupper(i))) {
        Size += sst.getStat()[i];
      }
   }
   return *this;
}
/** No descriptions */
const unsigned long* CSeqStat::getStat() const{
   return Stat;
}
/** No descriptions */
unsigned long CSeqStat::getLength() const{
   if(guessType() == NA) {
      if(sbWithN) {
        return Size;
      }
      return Size - Stat['X'] - Stat['N'];
   }
   else {
      if(sbWithN) {
        return Size + Stat[STOPCODON];
      }
      return Size - Stat['X'] + Stat[STOPCODON];
   }
}
};
