#include "segment.h"
#include <iostream>

//--------------------------------------
void segment::assign(length_t dFrom, length_t dTo) {
   m_dFrom = min(dFrom, dTo);
   m_dTo = max(dFrom, dTo);
}
//--------------------------------------
length_t segment::getLength() const {
   return m_dTo - m_dFrom + 1;
}
//--------------------------------------
length_t segment::getFrom() const {
   return m_dFrom;
}
//--------------------------------------
length_t segment::getTo() const {
   return m_dTo;
}
//--------------------------------------
void segment::setFrom(length_t dFrom) {
   assign(dFrom, getTo());
}
//--------------------------------------
void segment::setTo(length_t dTo) {
   assign(getFrom(), dTo);
}
//--------------------------------------
segment& segment::erase() {
   setTo(0);
   setFrom(0);
   return *this;
}
//--------------------------------------
bool segment::overlaps(const segment& seg) const {
   if (max(getTo(), seg.getTo()) - min(getFrom(), seg.getFrom()) + 1 < getLength() + seg.getLength()) {
      return true;
   }
   return false;
}
//--------------------------------------
bool segment::contains(const segment& seg) const {
   if (getFrom() <= seg.getFrom() && getTo() >= seg.getTo()) {
      return true;
   }
   return false;
}
//--------------------------------------
bool segment::inside(const segment& seg) const {
   return seg.contains(*this);
}
//--------------------------------------
segment& segment::add(const segment& seg) {
   setFrom(min(getFrom(), seg.getFrom()));
   setTo(max(getTo(), seg.getTo()));
   return *this;
}
//--------------------------------------
segment& segment::sub(const segment& seg) {
   if (overlaps(seg)) {
      if (inside(seg)) {
         erase();
         return *this;
      }
      if (contains(seg)) {
         return *this;
      }
      if (getFrom() < seg.getFrom() && getTo() < seg.getTo()) {
         setTo(seg.getFrom() - 1);
      }
      if (getFrom() > seg.getFrom() && getTo() > seg.getTo()) {
         setFrom(seg.getTo() + 1);
      }
   }
   return *this;
}
//--------------------------------------
length_t segment::dist(const segment& seg) const {
   return max(getFrom(), seg.getFrom()) - min(getTo(), seg.getTo());
}
//--------------------------------------
void segment::print(FILE* FOUT) const {
   fprintf(FOUT, "(%d, %d)\n", getFrom(), getTo());
}
//--------------------------------------
void segment::print(ostream& os) const {
   os << "(" << getFrom() << ", " << getTo() << ")\n";
}
/** -------------------------------------- */
segment& segment::operator =(const segment& seg) {
   assign(seg.getFrom(), seg.getTo());
   return *this;
}
//--------------------------------------
ostream& operator <<(ostream& os, const segment& seg) {
   seg.print(os);
   return os;
}
//--------------------------------------
